source('Prior_funk.R') # contains functions used below

#---------------------------------------
# Basic properties of the case study
#---------------------------------------
# 1. Define number of controls, periods and MonteCarlo samples
ncontrol=3;nperiod=5;nsim=100000
# 2. Define which parameters are varying in parameterization (b1,a1,c1,b2,a2,c2,b3,a3,c3)
isVar=c(T,F,F,T,F,F,F,F,F)
# 3. Define parameter names and prior distributions for b,a,c (of any control)
names.bac=c('b','a','c')
margins.bac=c('Gaussian','LogNormal','Gaussian')

#---------------------------------------
# Define priors on "physical" parameters
#---------------------------------------
# Priors are defined through Monte Carlo samples
# 1. Priors for low-flow control (control 1)
b1<-rnorm(nsim,mean=-0.6,sd=0.5) # weir activation stage
c1<-rnorm(nsim,mean=1.5,sd=0.025) # exponent
Bw1<-rlnorm(nsim,meanlog=log(8),sdlog=0.25) # weir width
Cr1<-rlnorm(nsim,meanlog=log(0.4),sdlog=0.25) # Discharge coefficient
g1<-rlnorm(nsim,meanlog=log(9.81),sdlog=0.0005) # gravity
# 2. Priors for main channel (control 2)
b2<-rnorm(nsim,mean=-0.6,sd=0.5) # channel bed stage
c2<-rnorm(nsim,mean=5/3,sd=0.025) # exponent
Bc2<-rlnorm(nsim,meanlog=log(15),sdlog=0.2) # channel width
KS2<-rlnorm(nsim,meanlog=log(25),sdlog=0.1) # Strickler coefficient
S02<-rlnorm(nsim,meanlog=log(0.005),sdlog=0.5) # slope
# 3. Priors for floodway (control 3)
b3<-rnorm(nsim,mean=1.2,sd=0.2) # channel bed stage
c3<-rnorm(nsim,mean=5/3,sd=0.025) # exponent
Bc3<-rlnorm(nsim,meanlog=log(30),sdlog=0.2) # channel width
KS3<-rlnorm(nsim,meanlog=log(15),sdlog=0.2) # Strickler coefficient
S03<-rlnorm(nsim,meanlog=log(0.005),sdlog=0.5) # slope
# 4. Priors for incremental global changes
d.g<-matrix(rnorm(nsim*(nperiod-1),mean=0,sd=0.5),nrow=nsim,ncol=nperiod-1)
# 5. Priors for incremental local changes
d.l<-matrix(rnorm(nsim*(nperiod-1),mean=0,sd=0.5),nrow=nsim,ncol=nperiod-1)
# 6. Starting point for all parameters
start=list(
  b1=-0.6,c1=1.5,Bw1=8,Cr1=0.4,g1=9.81,  # control 1
  b2=-0.6,c2=5/3,Bc2=15,KS2=25,S02=0.005, # control 2
  b3=1.2,c3=5/3,Bc3=30,KS3=15,S03=0.005,  # control 3
  d.l=rep(0,nperiod-1),     # incremental local changes d.l
  d.g=rep(0,nperiod-1)      # incremental global changes d.g
)

#---------------------------------------
# Perform Monte-Carlo propagation
#---------------------------------------
MC=propagate_Meyras(b1,c1,Bw1,Cr1,g1,  # control 1: rectangular weir
                    b2,c2,Bc2,KS2,S02, # control 2: large rectangular channel
                    b3,c3,Bc3,KS3,S03, # control 3: large rectangular channel
                    d.g,d.l,           # global and local incremental changes
                    start              # starting vector
                    )

#---------------------------------------
# Fit prior distribution on MC samples
#---------------------------------------
# define marginal prior distributions and parameter names
margins=c();names=c();k=0
for(i in 1:ncontrol){
  for(j in 1:3){
    k=k+1
    if(isVar[k]){
      margins=c(margins,rep(margins.bac[j],nperiod))
      names=c(names,paste(names.bac[j],i,'_',1:nperiod,sep=''))      
    }else{
      margins=c(margins,margins.bac[j])
      names=c(names,paste(names.bac[j],i,sep=''))
    }  
  }
}
# fit multivariate prior distribution
prior=fit(MC$sim,margins,names)
# write config files
colperiod=4 # columns where period will be stored in BaM data file
writeConfigFiles(prior,MC$start,ncontrol,nperiod,colperiod,isVar)
