propagate_Meyras<-function(b1,c1,Bw1,Cr1,g1,  # control 1: rectangular weir
                           b2,c2,Bc2,KS2,S02, # control 2: large rectangular channel
                           b3,c3,Bc3,KS3,S03, # control 3: large rectangular channel
                           d.g,d.l,           # global and local incremental changes
                           start              # starting vector
                           ){
  #^******************************************************************************
  #^* PURPOSE: Monte-Carlo propagation to move from "physical" parameterization P1
  #^*          to "inference" parameterization P2:
  #^*          P1=(b1,c1,Bw1,Cr1,g1,  # 1st control: offset, exponent, weir width, discharge coeff., gravity
  #^*              b2,c2,Bc2,KS2,S02, # 2nd control: offset, exponent, channel width, Strickler, slope
  #^*              b3,c3,Bc3,KS3,S03, # 3rd control: offset, exponent, channel width, Strickler, slope
  #^*              d.g,               # incremental overall changes affecting both the weir and the channel (size nperiod-1)
  #^*              d.l)               # incremental local changes affecting only the weir (size nperiod-1)
  #^*          P2=(b1,       # PERIOD-SPECIFIC: parameters b1 for all periods (size nperiod)
  #^*              a1,c1,    # STATIC: coefficient and exponent of 1st control
  #^*              b2,       # PERIOD-SPECIFIC: parameters b2 for all periods (size nperiod)
  #^*              a2,c2,    # STATIC: coefficient and exponent of 2nd control
  #^*              b3,a3,c3) # STATIC: all parameters of 3rd control
  #^******************************************************************************
  #^* PROGRAMMERS: Benjamin Renard, Irstea Lyon
  #^******************************************************************************
  #^* CREATED/MODIFIED: 09/01/2018
  #^******************************************************************************
  #^* IN
  #^*    1. [numeric vector] b1, weir activation stage (1st control)
  #^*    2. [numeric vector] c1, exponent for a rectangualar weir (1st control)
  #^*    3. [numeric vector] Bw1, weir width (1st control)
  #^*    4. [numeric vector] Cr1, discharge coefficienr for the weir (1st control)
  #^*    5. [numeric vector] g1, gravity (1st control)
  #^*    6. [numeric vector] b2, main channel bed stage (2nd control)
  #^*    7. [numeric vector] c2, exponent for a large rectangular channel (2nd control)
  #^*    8. [numeric vector] Bc2, channel width (2nd control)
  #^*    9. [numeric vector] KS2, Strickler coefficient (2nd control)
  #^*    10. [numeric vector] S02, Channel slope (2nd control)
  #^*    11. [numeric vector] b3, floodway channel bed stage (3rd control)
  #^*    12. [numeric vector] c3, exponent for a large rectangular channel (3rd control)
  #^*    13. [numeric vector] Bc3, channel width (3rd control)
  #^*    14. [numeric vector] KS3, Strickler coefficient (3rd control)
  #^*    15. [numeric vector] S03, Channel slope (3rd control)
  #^*    16. [numeric matrix] d.g, incremental global changes; size nsim*(nperiod-1)
  #^*    17. [numeric matrix] d.l, incremental local changes; size nsim*(nperiod-1)
  #^*    18. [list] start, starting vector for all parameters above
  #^* OUT
  #^*    A list containing:
  #^*    1. [data.frame] sim: Monte-Carlo samples in "inference" parameterization P2   
  #^*    2. [numeric vector] start: starting point in "inference" parameterization P2   
  #^******************************************************************************
  #^* REF.: Mansanarez et al. (2019) Shift happens! Adjusting stage-discharge 
  #^*       rating curves to morphological changes at known times. Water
  #^*       Resources Research.
  #^******************************************************************************
  #^* TO DO: /
  #^******************************************************************************
  #^* COMMENTS: /
  #^******************************************************************************
  
  nsim=length(b1);nperiod=NCOL(d.g)+1
  # initialise parameters that will be computed by propagation
  a1=vector(mode='double',length=nsim)
  a2=vector(mode='double',length=nsim)
  a3=vector(mode='double',length=nsim)
  allb1=matrix(NA,nsim,nperiod)
  allb2=matrix(NA,nsim,nperiod)
  #------------------------------
  # Monte-Carlo propagation
  #------------------------------
  for(i in 1:nsim){
    # Compute a1, a2 and a3
    a1[i]=Cr1[i]*Bw1[i]*sqrt(2*g1[i])
    a2[i]=KS2[i]*Bc2[i]*sqrt(S02[i])
    a3[i]=KS3[i]*Bc3[i]*sqrt(S03[i])
    # compute b1 and b2 for all periods
    allb1[i,1]=b1[i]
    allb1[i,2:nperiod]=b1[i]-cumsum(d.g[i,]+d.l[i,])
    allb2[i,1]=b2[i]
    allb2[i,2:nperiod]=b2[i]-cumsum(d.g[i,])
  }
  #-------------------------------------------------
  # transform starting vector in parameterization P2
  #-------------------------------------------------
  allb1.s=vector(mode='double',length=nperiod)
  allb2.s=vector(mode='double',length=nperiod)
  # All b1's
  allb1.s[1]=start$b1
  allb1.s[2:nperiod]=start$b1-cumsum(start$d.g+start$d.l)
  # All b2's
  allb2.s[1]=start$b2
  allb2.s[2:nperiod]=start$b2-cumsum(start$d.g)
  # a's
  a1.s=start$Cr1*start$Bw1*sqrt(2*start$g1)
  a2.s=start$KS2*start$Bc2*sqrt(start$S02)
  a3.s=start$KS3*start$Bc3*sqrt(start$S03)
  # return results
  return(list(
    sim=data.frame(b1=allb1,a1=a1,c1=c1,b2=allb2,a2=a2,c2=c2,b3=b3,a3=a3,c3=c3),
    start=c(allb1.s,a1.s,start$c1,allb2.s,a2.s,start$c2,start$b3,a3.s,start$c3)
    )
  )
}

propagate_Wairau<-function(b1,c1,Bc1,KS1,S01, # control 1: low-flow channel
                           b2,c2,Bc2,KS2,S02, # control 2: high-flow channel
                           d.b1,d.Bc1,        # incremental changes on low-flow channel bed and width
                           start              # starting vector
){
  #^******************************************************************************
  #^* PURPOSE: Monte-Carlo propagation to move from "physical" parameterization P1
  #^*          to "inference" parameterization P2:
  #^*          P1=(b1,c1,Bc1,KS1,S01, # 1st control: offset, exponent, channel width, Strickler, slope
  #^*              b2,c2,Bc2,KS2,S02, # 2nd control: offset, exponent, channel width, Strickler, slope
  #^*              d.b1,              # incremental changes on low-flow channel bed (size nperiod-1)
  #^*              d.Bc1)             # incremental changes on low-flow channel width (size nperiod-1)
  #^*          P2=(b1,       # PERIOD-SPECIFIC: parameters b1 for all periods (size nperiod)
  #^*              a1,       # PERIOD-SPECIFIC: parameters a1 for all periods (size nperiod)
  #^*              c1,       # STATIC: exponent of 1st control
  #^*              b2,a2,c2) # STATIC: all parameters of 2nd control
  #^******************************************************************************
  #^* PROGRAMMERS: Benjamin Renard, Irstea Lyon
  #^******************************************************************************
  #^* CREATED/MODIFIED: 16/01/2018
  #^******************************************************************************
  #^* IN
  #^*    1. [numeric vector] b1, low-flow channel bed stage (1st control)
  #^*    2. [numeric vector] c1, exponent for a large rectangular channel (1st control)
  #^*    3. [numeric vector] Bc1, channel width (1st control)
  #^*    4. [numeric vector] KS1, Strickler coefficient (1st control)
  #^*    5. [numeric vector] S01, Channel slope (1st control)
  #^*    6. [numeric vector] b2, high-flow channel bed stage (2nd control)
  #^*    7. [numeric vector] c2, exponent for a large rectangular channel (2nd control)
  #^*    8. [numeric vector] Bc2, channel width (2nd control)
  #^*    9. [numeric vector] KS2, Strickler coefficient (2nd control)
  #^*    10. [numeric vector] S02, Channel slope (2nd control)
  #^*    11. [numeric matrix] d.b1, incremental changes on low-flow channel bed; size nsim*(nperiod-1)
  #^*    12. [numeric matrix] d.Bc1, incremental changes on low-flow channel width; size nsim*(nperiod-1)
  #^*    13. [list] start, starting vector for all parameters above
  #^* OUT
  #^*    A list containing:
  #^*    1. [data.frame] sim: Monte-Carlo samples in "inference" parameterization P2   
  #^*    2. [numeric vector] start: starting point in "inference" parameterization P2   
  #^******************************************************************************
  #^* REF.: Mansanarez et al. (2019) Shift happens! Adjusting stage-discharge 
  #^*       rating curves to morphological changes at known times. Water
  #^*       Resources Research.
  #^******************************************************************************
  #^* TO DO: /
  #^******************************************************************************
  #^* COMMENTS: /
  #^******************************************************************************
  
  nsim=length(b1);nperiod=NCOL(d.b1)+1
  #------------------------------
  # Monte-Carlo propagation
  #------------------------------
  a2=vector(mode='double',length=nsim)
  allb1=matrix(NA,nsim,nperiod)
  alla1=matrix(NA,nsim,nperiod)
  # Start Monte-Carlo propagation
  for(i in 1:nsim){
    # Compute a2
    a2[i]=KS2[i]*Bc2[i]*sqrt(S02[i])
    # compute b1 and a1 for all periods
    allb1[i,1]=b1[i]
    allb1[i,2:nperiod]=b1[i]-cumsum(d.b1[i,])
    alla1[i,1]=KS1[i]*Bc1[i]*sqrt(S01[i])
    alla1[i,2:nperiod]=alla1[i,1]/cumprod(d.Bc1[i,])
  }
  #-------------------------------------------------
  # transform starting vector in parameterization P2
  #-------------------------------------------------
  allb1.s=vector(mode='double',length=nperiod)
  alla1.s=vector(mode='double',length=nperiod)
  # All b1's
  allb1.s[1]=start$b1
  allb1.s[2:nperiod]=start$b1-cumsum(start$d.b1)
  # All a1's
  alla1.s[1]=start$KS1*start$Bc1*sqrt(start$S01)
  alla1.s[2:nperiod]=alla1.s[1]/cumprod(start$d.Bc1)
  # a2
  a2.s=start$KS2*start$Bc2*sqrt(start$S02)
  
  # return results
  return(list(
    sim=data.frame(b1=allb1,a1=alla1,c1=c1,b2=b2,a2=a2,c2=c2),
    start=c(allb1.s,alla1.s,start$c1,start$b2,a2.s,start$c2)
  )
  )
}

propagate_Wairau_V2<-function(b1,c1,Bc1,KS1,S01, # control 1: low-flow channel
                              b2,c2,Bc2,KS2,S02, # control 2: high-flow channel
                              d.b1,d.Bc1,d.b2,   # incremental changes on low-flow channel bed and width and high-flow channel bed
                              start              # starting vector
){
  #^******************************************************************************
  #^* PURPOSE: Monte-Carlo propagation to move from "physical" parameterization P1
  #^*          to "inference" parameterization P2:
  #^*          P1=(b1,c1,Bc1,KS1,S01, # 1st control: offset, exponent, channel width, Strickler, slope
  #^*              b2,c2,Bc2,KS2,S02, # 2nd control: offset, exponent, channel width, Strickler, slope
  #^*              d.b1,              # incremental changes on low-flow channel bed (size nperiod-1)
  #^*              d.Bc1,             # incremental changes on low-flow channel width (size nperiod-1)
  #^*              d.b2)              # incremental changes on high-flow channel bed (size nperiod-1)
  #^*          P2=(b1,       # PERIOD-SPECIFIC: parameters b1 for all periods (size nperiod)
  #^*              a1,       # PERIOD-SPECIFIC: parameters a1 for all periods (size nperiod)
  #^*              c1,       # STATIC: exponent of 1st control
  #^*              b2,       # PERIOD-SPECIFIC: parameters b2 for all periods (size nperiod)
  #^*              a2,c2)    # STATIC: a/c parameters of 2nd control
  #^******************************************************************************
  #^* PROGRAMMERS: Benjamin Renard, Irstea Lyon
  #^******************************************************************************
  #^* CREATED/MODIFIED: 16/01/2018
  #^******************************************************************************
  #^* IN
  #^*    1. [numeric vector] b1, low-flow channel bed stage (1st control)
  #^*    2. [numeric vector] c1, exponent for a large rectangular channel (1st control)
  #^*    3. [numeric vector] Bc1, channel width (1st control)
  #^*    4. [numeric vector] KS1, Strickler coefficient (1st control)
  #^*    5. [numeric vector] S01, Channel slope (1st control)
  #^*    6. [numeric vector] b2, high-flow channel bed stage (2nd control)
  #^*    7. [numeric vector] c2, exponent for a large rectangular channel (2nd control)
  #^*    8. [numeric vector] Bc2, channel width (2nd control)
  #^*    9. [numeric vector] KS2, Strickler coefficient (2nd control)
  #^*    10. [numeric vector] S02, Channel slope (2nd control)
  #^*    11. [numeric matrix] d.b1, incremental changes on low-flow channel bed; size nsim*(nperiod-1)
  #^*    12. [numeric matrix] d.Bc1, incremental changes on low-flow channel width; size nsim*(nperiod-1)
  #^*    13. [numeric matrix] d.b2, incremental changes on high-flow channel bed; size nsim*(nperiod-1)
  #^*    14. [list] start, starting vector for all parameters above
  #^* OUT
  #^*    A list containing:
  #^*    1. [data.frame] sim: Monte-Carlo samples in "inference" parameterization P2   
  #^*    2. [numeric vector] start: starting point in "inference" parameterization P2   
  #^******************************************************************************
  #^* REF.: Mansanarez et al. (2019) Shift happens! Adjusting stage-discharge 
  #^*       rating curves to morphological changes at known times. Water
  #^*       Resources Research.
  #^******************************************************************************
  #^* TO DO: /
  #^******************************************************************************
  #^* COMMENTS: /
  #^******************************************************************************
  
  nsim=length(b1);nperiod=NCOL(d.b1)+1
  #------------------------------
  # Monte-Carlo propagation
  #------------------------------
  a2=vector(mode='double',length=nsim)
  allb1=matrix(NA,nsim,nperiod)
  alla1=matrix(NA,nsim,nperiod)
  allb2=matrix(NA,nsim,nperiod)
  # Start Monte-Carlo propagation
  for(i in 1:nsim){
    # Compute a2
    a2[i]=KS2[i]*Bc2[i]*sqrt(S02[i])
    # compute b1, a1 and b2 for all periods
    allb1[i,1]=b1[i]
    allb1[i,2:nperiod]=b1[i]-cumsum(d.b1[i,])
    alla1[i,1]=KS1[i]*Bc1[i]*sqrt(S01[i])
    alla1[i,2:nperiod]=alla1[i,1]/cumprod(d.Bc1[i,])
    allb2[i,1]=b2[i]
    allb2[i,2:nperiod]=b2[i]-cumsum(d.b2[i,])
  }
  #-------------------------------------------------
  # transform starting vector in parameterization P2
  #-------------------------------------------------
  allb1.s=vector(mode='double',length=nperiod)
  alla1.s=vector(mode='double',length=nperiod)
  allb2.s=vector(mode='double',length=nperiod)
  # All b1's
  allb1.s[1]=start$b1
  allb1.s[2:nperiod]=start$b1-cumsum(start$d.b1)
  # All a1's
  alla1.s[1]=start$KS1*start$Bc1*sqrt(start$S01)
  alla1.s[2:nperiod]=alla1.s[1]/cumprod(start$d.Bc1)
  # All b2's
  allb2.s[1]=start$b2
  allb2.s[2:nperiod]=start$b2-cumsum(start$d.b2)
  # a2
  a2.s=start$KS2*start$Bc2*sqrt(start$S02)
  
  # return results
  return(list(
    sim=data.frame(b1=allb1,a1=alla1,c1=c1,b2=allb2,a2=a2,c2=c2),
    start=c(allb1.s,alla1.s,start$c1,allb2.s,a2.s,start$c2)
  )
  )
}

propagate_Rakaia<-function(b1,c1,Bc1,KS1,S01, # control 1: low-flow channel
                           b2,c2,Bc2,KS2,S02, # control 2: high-flow channel
                           d.b1,d.Bc1,        # incremental changes on low-flow channel bed and width
                           start              # starting vector
                           ){
  #^******************************************************************************
  #^* PURPOSE: Monte-Carlo propagation to move from "physical" parameterization P1
  #^*          to "inference" parameterization P2:
  #^*          P1=(b1,c1,Bc1,KS1,S01, # 1st control: offset, exponent, channel width, Strickler, slope
  #^*              b2,c2,Bc2,KS2,S02, # 2nd control: offset, exponent, channel width, Strickler, slope
  #^*              d.b1,              # incremental changes on low-flow channel bed (size nperiod-1)
  #^*              d.Bc1)             # incremental changes on low-flow channel width (size nperiod-1)
  #^*          P2=(b1,       # PERIOD-SPECIFIC: parameters b1 for all periods (size nperiod)
  #^*              a1,       # PERIOD-SPECIFIC: parameters a1 for all periods (size nperiod)
  #^*              c1,       # STATIC: exponent of 1st control
  #^*              b2,a2,c2) # STATIC: all parameters of 2nd control
  #^******************************************************************************
  #^* PROGRAMMERS: Benjamin Renard, Irstea Lyon
  #^******************************************************************************
  #^* CREATED/MODIFIED: 16/01/2018
  #^******************************************************************************
  #^* IN
  #^*    1. [numeric vector] b1, low-flow channel bed stage (1st control)
  #^*    2. [numeric vector] c1, exponent for a large rectangular channel (1st control)
  #^*    3. [numeric vector] Bc1, channel width (1st control)
  #^*    4. [numeric vector] KS1, Strickler coefficient (1st control)
  #^*    5. [numeric vector] S01, Channel slope (1st control)
  #^*    6. [numeric vector] b2, high-flow channel bed stage (2nd control)
  #^*    7. [numeric vector] c2, exponent for a large rectangular channel (2nd control)
  #^*    8. [numeric vector] Bc2, channel width (2nd control)
  #^*    9. [numeric vector] KS2, Strickler coefficient (2nd control)
  #^*    10. [numeric vector] S02, Channel slope (2nd control)
  #^*    11. [numeric matrix] d.b1, incremental changes on low-flow channel bed; size nsim*(nperiod-1)
  #^*    12. [numeric matrix] d.Bc1, incremental changes on low-flow channel width; size nsim*(nperiod-1)
  #^*    13. [list] start, starting vector for all parameters above
  #^* OUT
  #^*    A list containing:
  #^*    1. [data.frame] sim: Monte-Carlo samples in "inference" parameterization P2   
  #^*    2. [numeric vector] start: starting point in "inference" parameterization P2   
  #^******************************************************************************
  #^* REF.: Mansanarez et al. (2019) Shift happens! Adjusting stage-discharge 
  #^*       rating curves to morphological changes at known times. Water
  #^*       Resources Research.
  #^******************************************************************************
  #^* TO DO: /
  #^******************************************************************************
  #^* COMMENTS: /
  #^******************************************************************************
  
  nsim=length(b1);nperiod=NCOL(d.b1)+1
  #------------------------------
  # Monte-Carlo propagation
  #------------------------------
  a2=vector(mode='double',length=nsim)
  allb1=matrix(NA,nsim,nperiod)
  alla1=matrix(NA,nsim,nperiod)
  # Start Monte-Carlo propagation
  for(i in 1:nsim){
    # Compute a2
    a2[i]=KS2[i]*Bc2[i]*sqrt(S02[i])
    # compute b1 and a1 for all periods
    allb1[i,1]=b1[i]
    allb1[i,2:nperiod]=b1[i]-cumsum(d.b1[i,])
    alla1[i,1]=KS1[i]*Bc1[i]*sqrt(S01[i])
    alla1[i,2:nperiod]=alla1[i,1]/cumprod(d.Bc1[i,])
  }
  #-------------------------------------------------
  # transform starting vector in parameterization P2
  #-------------------------------------------------
  allb1.s=vector(mode='double',length=nperiod)
  alla1.s=vector(mode='double',length=nperiod)
  # All b1's
  allb1.s[1]=start$b1
  allb1.s[2:nperiod]=start$b1-cumsum(start$d.b1)
  # All a1's
  alla1.s[1]=start$KS1*start$Bc1*sqrt(start$S01)
  alla1.s[2:nperiod]=alla1.s[1]/cumprod(start$d.Bc1)
  # a2
  a2.s=start$KS2*start$Bc2*sqrt(start$S02)
  
  # return results
  return(list(
    sim=data.frame(b1=allb1,a1=alla1,c1=c1,b2=b2,a2=a2,c2=c2),
    start=c(allb1.s,alla1.s,start$c1,start$b2,a2.s,start$c2)
    )
  )
}


fit<-function(sim, # Monte Carlo simulations produced by function propagate_XXX$sim
              margins=rep('Gaussian',NCOL(sim)), # marginal distributions
              names=paste('par',1:NCOL(sim),sep=''), # parameter names
              ncol=7,rowmax=4 # graphical parameters (here 7*4 parameters shown per figure)
              ){
  #^******************************************************************************
  #^* PURPOSE: fit a multivariate prior distribution to Monte-Carlo samples
  #^******************************************************************************
  #^* PROGRAMMERS: Benjamin Renard, Irstea Lyon
  #^******************************************************************************
  #^* CREATED/MODIFIED: 16/01/2018
  #^******************************************************************************
  #^* IN
  #^*    1. [numeric matrix] sim, Monte Carlo simulations (nsim*npar)
  #^*    2. [character vector] margins, marginal distributions (npar)
  #^*    3. [character vector] names, parameter names (npar)
  #^*    4. [integer] ncol, rowmax: graphical parameters controlling the number of subplots per figure
  #^* OUT
  #^*    A list containing:
  #^*    1. [character vector] margins, marginal distributions (same as the one used as input)
  #^*    2. [list of vectors] priorpar, prior parameters for each marginal prior    
  #^*    3. [numeric vector] start, starting points (e.g. prior mean)    
  #^*    4. [numeric matrix] cor, prior correlation (in Gaussian space)    
  #^******************************************************************************
  #^* REF.: Mansanarez et al. (2019) Shift happens! Adjusting stage-discharge 
  #^*       rating curves to morphological changes at known times. Water
  #^*       Resources Research.
  #^******************************************************************************
  #^* COMMENTS: only normal and lognormal margins available for the moment
  #^******************************************************************************
  #^* TO DO: can probably be generalized to any marginal distribution
  #^******************************************************************************
  
  # check sizes match and if so, initialize
  n=NCOL(sim)
  if(length(margins)!=n){
    message('Fatal:size mismatch [sim,margins]');return(NA)
  }
  priorpar=vector("list",n)
  transform=matrix(NA,nsim,n)
  # start computations for each margin
  for(i in 1:n){
    # marginal prior parameters
    priorpar[[i]]=switch(margins[i],
               'Gaussian'={c(mean(sim[,i]),sd(sim[,i]))},
               'LogNormal'={c(mean(log(sim[,i])),sd(log(sim[,i])))},
               NA)
    # Transform simulations into Gaussian space to estimate the correlation of the Gaussian copula
    transform[,i]=switch(margins[i],
                    'Gaussian'={qnorm(pnorm(sim[,i],mean=priorpar[[i]][1],sd=priorpar[[i]][2]))},
                    'LogNormal'={qnorm(plnorm(sim[,i],meanlog=priorpar[[i]][1],sdlog=priorpar[[i]][2]))},
                    NA)
  }
  # prior correlation
  corel=cor(transform)
  # Verify that marginal fit is acceptable graphically
  nrow=min(ceiling(n/ncol),rowmax)
  for(i in 1:n){
    if(i%%(nrow*ncol)==1) {X11();par(mfrow=c(nrow,ncol))}
    x=seq(min(sim[,i]),max(sim[,i]),,100)
    y=switch(margins[i],
             'Gaussian'={dnorm(x,mean=priorpar[[i]][1],sd=priorpar[[i]][2])},
             'LogNormal'={dlnorm(x,meanlog=priorpar[[i]][1],sdlog=priorpar[[i]][2])},
             NA)
    hist(sim[,i],breaks=50,freq=F,main=NULL,xlab=names[i],ylab='density')
    lines(x,y,col='red')
  }
  # show correlation matrix
  X11();image(1:n,1:n,corel)
  # Return
  return(list(margins=margins,priorpar=priorpar,cor=corel))
}

writeConfigFiles<-function(prior, # prior object produced by function 'fit'
                           start, # starting vector produced by function propagate_XXX$start
                           ncontrol, # number of hydraulic controls
                           nperiod,colperiod, # number of periods and column where it is stored in data file
                           isVar=rep(F,ncontrol*nperiod), # flag varying (period-specific) parameters
                           model='Config_Model.txt', # Model configuration file
                           correl='PriorCorrelation.txt', #prior correlation file
                           names.bac=c('b','a','c') # base name of the 3 parameters for one control
                           ){
  #^******************************************************************************
  #^* PURPOSE: write configuration files used by BaM
  #^******************************************************************************
  #^* PROGRAMMERS: Benjamin Renard, Irstea Lyon
  #^******************************************************************************
  #^* CREATED/MODIFIED: 17/01/2018
  #^******************************************************************************
  #^* IN
  #^*    1. [list] prior, object produced by function 'fit'
  #^*    2. [numeric vector] start, starting vector produced by function propagate_XXX$start
  #^*    3. [integer] ncontrol, number of hydraulic controls
  #^*    4. [integer] nperiod, number of periods 
  #^*    5. [integer] colperiod, column where period is stored in data file
  #^*    6. [logical vector] isVar, flag varying (period-specific) parameters
  #^*    7. [character] model, Model configuration file
  #^*    8. [character] correl, prior correlation file
  #^*    9. [character vector]  base name of the 3 parameters for one control
  #^* OUT
  #^*    Nothing - just write files
  #^******************************************************************************
  #^* REF.: Mansanarez et al. (2019) Shift happens! Adjusting stage-discharge 
  #^*       rating curves to morphological changes at known times. Water
  #^*       Resources Research.
  #^******************************************************************************
  #^* COMMENTS: /
  #^******************************************************************************
  #^* TO DO: /
  #^******************************************************************************
  
  #---------------------------------------
  # Config_Model file
  #---------------------------------------
  # Start assembling values to be written in file and associated comments
  comment=c(
    'Model ID',
    'nX: number of input variables',
    'nY: number of output variables',
    'nPar: number of parameters theta'
  )
  val=list('BaRatinBAC',1,1,3*ncontrol)
  # specification for each parameter
  k=0;m=1
  for(i in 1:ncontrol){ # loop on each hydraulic control
    for(j in 1:length(names.bac)){ # loop on each b-a-c parameter of the control
      k=k+1
      # comments
      comment=c(comment,
                'Parameter name',
                'Initial guess',
                'Prior distribution',
                'Prior parameters')
      # parname
      pname=paste(names.bac[j],i,sep='')
      val=c(val,pname)
      if(!isVar[k]){ # Static parameter
        # starting point
        val=c(val,start[m])
        # prior distribution
        val=c(val,prior$margins[m])
        # prior parameters
        val=c(val,list(prior$priorpar[[m]]))
        m=m+1                
        } else { # period-specific parameter
        # starting point
        val=c(val,-666.666) # unused, can use a dummy number
        # prior distribution: use 'VAR'
        val=c(val,'VAR')
        # Name of the configuration file for this varying parameter
        fname=paste('Config_',pname,'_VAR.txt',sep='')
        val=c(val,fname)
        # Write this additional configuration file
        comment0=c(
          'Number of periods',
          'Column where period is written in data file'
        )
        val0=list(nperiod,colperiod)
        for(ii in 1:nperiod){ # one parameter block per period
          # comments
          comment0=c(comment0,
                    'Parameter name',
                    'Initial guess',
                    'Prior distribution',
                    'Prior parameters')
          # par name
          val0=c(val0,paste(pname,'_',ii,sep=''))
          # starting point
          val0=c(val0,start[m])
          # prior distribution
          val0=c(val0,prior$margins[m])
          # prior parameters
          val0=c(val0,list(prior$priorpar[[m]]))
          m=m+1
        }
        writeConfig(val0,comment0,getwd(),fname)
      }       
    }
  }  
  writeConfig(val,comment,getwd(),model)
  
  #---------------------------------------
  # Prior correlations
  #---------------------------------------
  # add 2 lines/columns, uncorrelated with all others, for structural uncertainty parameters gamma1 and gamma2
  n=NCOL(prior$cor)
  foo=matrix(0,n+2,n+2)
  foo[1:n,1:n]=prior$cor
  foo[n+1,n+1]=1
  foo[n+2,n+2]=1
  # write resulting matrix
  write(foo,file=correl,ncolumns=n+2)
}

writeConfig<-function(val,comment,dir,fname,addQuote=T){
  #^******************************************************************************
  #^* PURPOSE: Generic function to write a configuration file
  #^******************************************************************************
  #^* PROGRAMMERS: Benjamin Renard, Irstea Lyon
  #^******************************************************************************
  #^* CREATED/MODIFIED: 12/10/2018
  #^******************************************************************************
  #^* IN
  #^*    1. [list] val, list of values to be written in the configuration file 
  #^*    2. [string vector] comment, comments (same length as val) 
  #^*    3. [string] dir, destination directory 
  #^*    4. [string] fname, file name 
  #^*    5. [logical] addQuote, add double quotes to any string encountered in val? 
  #^* OUT
  #^*    1. nothing  
  #^******************************************************************************
  #^* REF.: 
  #^******************************************************************************
  #^* TO DO: 
  #^******************************************************************************
  #^* COMMENTS: Overwrites any pre-existing file. Directory created if needed
  #^******************************************************************************
  
  n=length(val)
  txt=vector("character",n)
  for (i in 1:n){
    foo=val[[i]]
    # add double quotes to strings
    if(is.character(foo) & addQuote){foo=addQuotes(foo)}
    # transform R logicals to Fortran logicals
    if(is.logical(foo)){foo=paste(ifelse(foo,'.true.','.false.'))}
    # stitch values and comment
    if(is.null(comment[i])){
      txt[i]=paste(foo,collapse=',')
    } else {
      txt[i]=paste(paste(foo,collapse=','),'!',comment[i])
    }
  }
  
  if(!dir.exists(dir)){dir.create(dir,recursive=T)}
  file=file.path(dir,fname)
  
  write.table(matrix(txt, ncol = 1, nrow = length(txt)), file = file,
              row.names = FALSE, col.names = FALSE, quote = FALSE)
}

addQuotes<-function(txt){
  #^******************************************************************************
  #^* PURPOSE: add double quotes to a string or to each element of a string vector
  #^******************************************************************************
  #^* PROGRAMMERS: Benjamin Renard, Irstea Lyon
  #^******************************************************************************
  #^* CREATED/MODIFIED: 12/10/2018
  #^******************************************************************************
  #^* IN
  #^*    1. [string vector] txt
  #^* OUT
  #^*    1. [string vector] double-quoted string vector 
  #^******************************************************************************
  #^* REF.: 
  #^******************************************************************************
  #^* TO DO: 
  #^******************************************************************************
  #^* COMMENTS: 
  #^******************************************************************************
  
  n=length(txt)
  out=txt
  for(i in 1:n){
    out[i]=paste('"',txt[i],'"',sep='')
  }
  return(out)
}

